#!/usr/bin/perl
#
# contest.pl - an email contest program
#
# copyright(C) 2005 Scott Mazur <scott.AT.littlefish.ca>

# GNU General Public License
# --------------------------------------------------------------------
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# requires:
#     Time::Local;
#     Data::Dumper;
#     Mail::Mailer;
#     DB_File;

my $ver = '0.10';

# Version 0.10
# - initial release

#
# syntax:
#
# used in ~/.forward file of contest receiver email (pipe mode)
#   contest.pl [-c config] [-l logfile] [-p homepath] switchUser
#
#    -c config    Load configfile to over-ride the default config file.
#                 This config in turn will be over-ridden by any contest config.
#                 If the file name does not contain a leading '/' it will
#                 be prefixed by the homePath.
#
#    -l logfile   Send all output to a logfile.  You probably want to use
#                 this in pipe mode to avoid filling your mail logs with
#                 contest output.
#                 If the file name does not contain a leading '/' it will
#                 be prefixed by the homePath.
#
#    -p homepath  Specify a directory for the user home.
#                 (mostly for virtual user with no real unix home)
#                 If the homepath does not contain a leading '/' it will
#                 be prefixed by the user's real homepath.
#
#    switchUser   The program will switch to this user (if possible).
#                 The switch user parameter triggers the program into pipe mode.
#
# The .forward file of the contest receiving user will contain a line of the form:
#
#      "|/usr/local/bin/contest.pl -l messages.log username"
#
#
# used from the command line (pipe mode)
#   contest.pl [-c config] [-l logfile] [-p homepath] [-u switchUser]
#              [-N contest#[,contest#...]] [-e] [-E] [-P]
#              [-t] [-C] [-D dumpfile] [-L loadfile]
#              [-R] -[W] [-n]
#
#    -c config     Same as pipe mode.
#    -l logfile    Same as pipe mode.
#    -p homepath   Same as pipe mode.
#
#    -u switchUser Same as pipe mode.  CAREFUL! If you don't use the -u option empty
#                  databases (and/or paths) may be created with the uid of the current
#                  user which the contest user may not have permission to use.
#
#    -N contest#   A comma separated list of Contest numbers to use with this command.
#                  If none is given, the program will attempt to find all available
#                  contests based on contest config files that are found in the
#                  homepath directory.  Each contest provided (or found) will be acted
#                  on with the same command options.
#
#    -e            List email addresses
#
#    -E            List entries
#
#    -P            List pending entries (waiting confirmation of email)
#
#    -t            Test mode (no updates will be reported but not actioned)
#
#    -C            Perform Database Consistency checks, rebuild missing indexes, etc.
#
#    -D dumpfile   Dump the database Contents into a text file.
#
#    -L loadfile   Load a text file into the databases (same format as dumpfile).
#                  A consistency check is automatically triggered after load.
#
#    -R            Re-test entries against contest rules.  Those that pass will have their
#                  status set to 'T', those that fail will get 'F'.
#
#    -W            Scan contest entries for winning entry/entries.  Winners will have their
#                  status set to 'W', losers will get 'L'.  Checks are done to ensure winners
#                  aren't declared before the contest completes.
#
#    -n            Send notify emails to losers (who requested it). Checks are done to ensure
#                  that losers aren't notified before the contest completes.

#TODO: use a generic key for the pending entry.  Want to Embed the contest number in the pending data
#TODO: global email database - one email can enter many contests without confirming each contest
#TODO: email admin interface to allow admin person to perform command mode operations through email
#TODO: Command option to generate PHP file that can be included on webpage to collect entries, update expire/anounce dates, etc.
#TODO: Make the 'notify me if I lose' a configurable parameter instead of hard coding 'notify'
#TODO: Winner notification isn't automated yet

use strict;
use warnings; # testing only
#use Time::Local;
#use Data::Dumper;
use Mail::Mailer;
#use DB_File;

$ENV{PATH} = "/bin:/usr/bin"; # limit PATH

#FIXME: testing only
my $textPrefix = '';
$textPrefix = 'test.';

#FIXME: testing only
my $defaultConfigFile = "/home/virtual/aholeintheworld.com/contest/contest/${textPrefix}etc.contest.conf";
#my $defaultConfigFile = '/etc/contest/contest.conf';

my $stringOptions = 'clpuDLN';
my $flagOptions = 'iCPEetRWn';
my $usage = <<EOF;
Usage: contest.pl [-c configfile] [-l logfile] [-p homepath] switchUser
       contest.pl [-c configfile] [-l logfile] [-p homepath] [-u switchUser]
                  [-N contest[,contest...]]
                  [-P] [-E] [-e]
                  [-C] [-D dumpfile] [-L loadfile] [-t]
                  [-R] -[W] [-n]
EOF
my @ignore = qw(root daemon postmaster mailer-daemon mailer);

my %defaultConfig = (
    magickey => 'magic',
    homePath => '',
    logFile => '',
    ignore => [qw(root daemon postmaster mailer-daemon mailer)],
    CONTESTNAME => 'My Email Contest',
    CONTESTEMAIL => 'contest@mydomain.com',
    CONTESTSOURCE => '',
    EXPIRE => '1969-12-31:18:00:00',
    ANNOUNCE => '1969-12-31:18:00:00',
    RESPONSETIME => 24,
    ENTRYLIMIT => 0,
    WINNERS => 1,
    PICKTYPE => 'random', # or 'first'
    PICKLOSERS => 'true',

# contest data
    dataTypes => ['name:text:Name', 'notify:checkbox:Notify me if I lose', 'answer:text:Answer'],
    dataRules => ['answer:i:test'],

# send notify for these events
    notifyChallenge  => '',
    notifyEntered  => '',
    notifyExpired  => '',
    notifyLimit  => '',
    notifyCorrupt => '',
    notifyBlacklist => '',
    notifyIgnore => '',

    subjectChallenge => 'Reply Required!',
    messageChallenge => "<DATE> <TIME>, Contest entry received for: <EMAIL>\n\nno challenge...",

    subjectEntered => 'Entry accepted! <CONTESTNAME>',
    messageEntered => 'Your entry to the <CONTESTNAME> has been accepted.',

    subjectExpired => '<CONTESTNAME> has Expired!',
    messageExpired => 'Sorry, the <CONTESTNAME> expired on <EXPIRE.DATE> at <EXPIRE.TIME> Central Standard Time.',

    subjectClosed => '<CONTESTNAME> has Closed!',
    messageClosed => 'Sorry, the <CONTESTNAME> is now closed to entries.',

    subjectLimit => 'Entry Refused! <CONTESTNAME>',
    messageLimit => 'Sorry, your entry has been refused because it exceeds the <CONTESTNAME> entry limit of <ENTRYLIMIT>.',

    subjectLoser => "Entry did not win, <CONTESTNAME>",
    messageLoser => 'Sorry, your entry submitted <DATE>, <TIME> was not selected as a winner',

    subjectWinner => 'You won! <CONTESTNAME>',
    messageWinner => 'Congratulations!  Your entry submitted <DATE>, <TIME> has been selected to be a winning entry.',
);

#### SUPPORTING FUNCTIONS ####
package Transaction;
use Time::Local;

$Data::Dumper::Indent = 0; # translate nl to \n, etc
$Data::Dumper::Useqq = 1; # no spaces
$Data::Dumper::Sortkeys = 1; # sort the hash keys

my $ruleTrueFlag = 'T';
my $winnerTrueFlag = 'W';
my $notifyTrueFlag = 'N';
my %altStatusFlags = (
    $ruleTrueFlag => 'F',
    F => $ruleTrueFlag,
    $winnerTrueFlag => 'L',
    L => $winnerTrueFlag,
    $notifyTrueFlag => '',
);

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my ($mode, $options, $user) = @_;

    my $self = {
            mode => $mode,
            options => $options,
            %defaultConfig,
            STATUS => '?',
            NOTIFYLOSUR => 0,
            'EXPIRE.SECONDS' => 0,
            'EXPIRE.DATE' => 'Wed, Dec 31, 1969',
            'EXPIRE.TIME' => '6 pm',
            'ANNOUNCE.SECONDS' => 0,
            'ANNOUNCE.DATE' => 'Wed, Dec 31, 1969',
            'ANNOUNCE.TIME' => '6 pm',
            ME => 'My Email Contest <contest@mydomain.com>',
            logFh => 0,
            dataCount => 0,
            CONTEST => 'unknown',
       };
    bless $self, $class;

    $self->sec2date(time());

    # generate default values
    $self->formatDataTypes($defaultConfig{dataTypes});

    $self->loadConfig($defaultConfigFile);

    $self->loadConfig($options->{c}) if ($options->{c});

    # borrowed from vacation.pl
    $user = $user || $ENV{'USER'} || $ENV{'LOGNAME'} || getlogin || (getpwuid($>))[0];

    #starting point for home Dir;
    $self->{uid} = $>;
    $self->{gid} = $);
    my ($uid, $gid, $home) = (getpwnam($user))[2,3,7];
    if ($uid) {
        if ($uid != $>) {
            $self->{uid} = $uid;
            $self->{gid} = $gid;
            $) = $gid; # switch first
            $< = $> if ($> == 0);
            $> = $uid;
            $ENV{'HOME'} = $home;
        }
    }
    $home = $ENV{'HOME'} if ($ENV{'HOME'});
    # guess real homedir under automounter
    $home = "/export$home" if ( -d "/export$home" );

    if ($self->{homePath}) {
        $home = $self->{homePath} =~ m!^/! ? $self->{homePath} : "$home/$self->{homePath}";
    }

    if ($options->{p}) {
        my $optionPath = $self->cleanPath($options->{p});
        $home = $optionPath =~ m!^/! ? $optionPath : "$home/$optionPath";
    }
    -d $home or $self->die("No home directory for user $user");

    ($home =~ /^(.+)$/) && ($home = $1);  # untaint $home...
    chdir $home or $self->die("Can't chdir to $home: $!");

    $self->{homePath} = $home;

    my $logFile = $options->{l} || $self->{logFile};
    if ($logFile) {
        $logFile = $self->path($logFile);

        open(my $fh, ">>", $logFile) or $self->die("Failed to open $logFile: $!");
        $self->{logFh} = $fh;
    }
    $self->writeLog('Begin Transaction') if ($self->{mode} == 0);

    return $self;
}

sub path { my $self = shift;
    $_ = $self->cleanPath(shift);
#FIXME: testing only
    return (m!^/! ? $_ : $self->{homePath} . "/$textPrefix" . $_);
    return (m!^/! ? $_ : $self->{homePath} . '/' . $_);
}

sub cleanPath { my $self = shift;
    $_ = shift;
    s![?*]!.!g;
    s!/+!/!g;
    s!/*$!!;
    s!^[.]+/!!;
    return $_;
}

sub cleanup { my $self = shift;
    $self->writeLog('End Transaction') if ($self->{mode} == 0);
    close($self->{logFh}) if ($self->{logFh});
}

sub die { my $self = shift;
    my $txt = shift;
    $self->writeLog("Transaction died: $txt") if ($self->{logFh});
    die "$txt\n";
}

sub contestPath { my $self = shift;
    my $ext = shift;
    return $self->path('contest.' . $self->{CONTEST} . '.' . $ext);
}

sub writeLog { my $self = shift;
    my $txt = shift;
    my $time = scalar(localtime);
    $time =~ s/^\S+\s+(\S+)\s+(\S+)\s+(\S+).*/$1 $2 $3 - /;

    if (my $fh = $self->{logFh}) {
        foreach (split(/\n/, $txt)) {
            print $fh "$time $txt\n";
        }
    } else {
        print "$txt\n";
    }
}

sub writeContestLog { my $self = shift;
    my ($status, $msg) = @_;
    return $self->writeLog("Contest #$self->{CONTEST} - ". uc($status) . " $msg");
}

sub sec2date { my $self = shift;
    my ($seconds, $prefix) = @_;
    $prefix = $prefix ? "$prefix." : '';;
    $self->{$prefix.'SECONDS'} = $seconds;
    my $dateStr = localtime($seconds);
    if ($dateStr =~ /^(\S+) (\S+)\s+(\d+) (\d\d):(\d\d):(\d\d) (\d\d\d\d)/) {
        my ($day, $month, $date, $hour, $min, $secs, $year) = ($1, $2, $3, $4, $5, $6, $7);
        my $am = $hour > 11 ? 'pm' : 'am';
        $hour -= 12 if ($hour > 12);
        $hour .= ":$min" if (($min > 0) or ($secs> 0));
        $hour .= ":$secs" unless $secs == 0;
        $self->{$prefix.'DATE'} = "$day, $month $date, $year";
        $self->{$prefix.'TIME'} = "$hour $am";
        return 1;
    }
    $self->{$prefix.'DATE'} = '?';
    $self->{$prefix.'TIME'} = '?';
    return 0;
}

sub setStatusFlags { my $self = shift;
    my $flag = shift;
    my $status = $self->{STATUS};
    my $test = "?$flag$altStatusFlags{$flag}";
    my $prev = ($status =~ m/([$test]+)/) ? $1 : '?';
    return 0 if ($prev eq $flag); # no change
    # remove any previous flags
    $status =~ s/[$test]+//g;
    $self->{STATUS} = $status.$flag;
    return $prev;
}

sub isTrue { my $self = shift;
    my $status = shift || $self->{STATUS};
    return ($status =~ m/$ruleTrueFlag/);
}
sub isWinner { my $self = shift;
    my $status = shift || $self->{STATUS};
    return ($status =~ m/$winnerTrueFlag/);
}
sub setWinner { my $self = shift;
    return $self->setStatusFlags($winnerTrueFlag);
}
sub setLoser { my $self = shift;
    return $self->setStatusFlags($winnerTrueFlag);
}
sub isNotified { my $self = shift;
    my $status = shift || $self->{STATUS};
    return ($status =~ m/$notifyTrueFlag/);
}
sub setNotified { my $self = shift;
    return $self->setStatusFlags($notifyTrueFlag);
}
#2005-04-01:13:13:00
sub formatDate { my $self = shift;
    my ($param, $valuePtr) = @_;
    if ($valuePtr) {
        $self->{$param} = $$valuePtr ? $$valuePtr : '1969-12-31:18:00:00';
    }
    if ($self->{$param} =~ m/^(\d\d\d\d)-(\d\d)-(\d\d):(\d\d):(\d\d):(\d\d)/) {
        return $self->sec2date(timelocal($6, $5, $4, $3, $2 - 1, $1 - 1900), $param);
    }
    return 0;
}

sub formatDataTypes { my $self = shift;
    my $arrayPtr = shift;
    foreach (@$arrayPtr) {
        if (m/^\s*(\w+)\s*:\s*(text|checkbox)\s*:\s*(.*)$/i) {
            my $k = uc($1);
            $self->{dataCount}++ unless ($self->{dataTypes_hash}->{$k});
            $self->{dataTypes_hash}->{$k} = $2;
            $self->{"RESULTS.$k.DESC"} = $3;
        } else {
            $self->configError("format for dataTypes is 'label:(text|checkbox):Description' but found '$_'");
        }
    }
}

sub formatDataRules { my $self = shift;
    my $arrayPtr = shift;
    foreach (@$arrayPtr) {
        if (/^(\w+):((and|or|!|i)*):(.+)/) {
            push @{$self->{dataRules}}, join(':', $1,$2,$4);
        } else {
            $self->configError("format for dataRules is 'label:[and|or|!|i]...:regex' but found '$_'");
        }
    }
}

sub ruleTable { my $self = shift;
    my @rules = ();
    my $first = 1;
    foreach (@{$self->{dataRules}}) {
        next unless (/^([^:]+):([^:]*):(.*)/);
        my ($label, $cond, $regex) = (uc($1), $2, $3);
        if ($first) {
            $cond =~ s/(and|or)//g;
            $first = 0;
        }
        push @rules, join(':', $label, $cond, $regex);
    }
    return \@rules;
}

sub testRules { my $self = shift;
    my $rulesPtr = shift;
    my $result = 1;
    foreach (@$rulesPtr) {
        my ($label, $cond, $regex) = split(/:/);
        $label = "RESULTS.$label" if $self->{"RESULTS.$label"};
        next unless $self->{$label};
        return $altStatusFlags{$ruleTrueFlag} unless ($result or $cond =~ /or/);
        return $ruleTrueFlag if ($result and $cond =~ /or/);
        if ($cond =~ /!/) {
            if ($cond =~ /i/) {
                $result = ($self->{$label} !~ /$regex/i);
            } else {
                $result = ($self->{$label} !~ /$regex/);
            }
        } elsif ($cond =~ /i/) {
            $result = ($self->{$label} =~ /$regex/i);
        } else {
            $result = ($self->{$label} =~ /$regex/);
        }
    }
    return ($result ? $ruleTrueFlag : $altStatusFlags{$ruleTrueFlag});
}

# eliminate duplicate entries
sub formatIgnore { my $self = shift;
    my $arrayPtr = shift;
    my %hash = ();
    foreach (@$arrayPtr) {
        $hash{lc($_)} = 1;
    }
    $self->{ignore} = [keys %hash];
}

sub configError { my $self = shift;
    my $msg = shift;
    $self->die("Config Error $self->{configFile}, line $self->{configKeyLine}: $msg");
}

sub cleanArray { my $self = shift;
    my $valuePtr = shift;

    my @results = ();
    while ($$valuePtr) {
        $$valuePtr =~ s/^([^\s,"']*)([\s,"']?)//;
        my ($val, $sep) = ($1, $2);
        if ($sep =~ /['"]/) {
            $$valuePtr =~ s/^([^$sep]*)($sep?)//;
            $val .= $1;
        }
        next if ($val eq '');
        push @results, $val;
    }

    return \@results;
}

# validate config values
sub applyConfig { my $self = shift;
    my ($keyPtr, $valuePtr) = @_;

    $self->configError("invalid parameter $$keyPtr") unless exists $defaultConfig{$$keyPtr};

    $$valuePtr = '' if ($$valuePtr =~ /^none$/i);

    if ($$keyPtr eq 'dataTypes') {
        $self->formatDataTypes($self->cleanArray($valuePtr));
    } elsif ($$keyPtr eq 'dataRules') {
        $self->{dataRules} = [];
        $self->formatDataRules($self->cleanArray($valuePtr));
    } elsif ($$keyPtr eq 'ignore') {
        $self->formatIgnore($self->cleanArray($valuePtr));
    } elsif ($$keyPtr =~ m/(CONTESTEMAIL|notify)/) { # email
        #FIXME: edit email format
        $self->{$$keyPtr} = $$valuePtr;
    } elsif ($$keyPtr =~ m/message/) { # substitution strings
        $$valuePtr =~ s/\\n/\n/g;
        $$valuePtr =~ s/\\t/\t/g;
        $self->{$$keyPtr} = $$valuePtr;
    } elsif ($$keyPtr =~ m/(homePath|logFile)/) { # file paths
        $self->{$$keyPtr} = $self->cleanPath($$valuePtr);
    } elsif ($$keyPtr =~ m/(EXPIRE|ANNOUNCE)/) { # formatted dates
        $self->configError("format for $$keyPtr is 'YYYY-MM-DD:HH:MM:SS'") unless $self->formatDate($$keyPtr, $valuePtr);
    } elsif ($$keyPtr =~ m/(RESPONSETIME|ENTRYLIMIT|WINNERS)/) { # integer
        $self->configError("$$keyPtr must be an integer value") unless $$valuePtr =~ /^\d*$/;
        $self->{$$keyPtr} = $$valuePtr + 0;
    } elsif ($$keyPtr =~ m/(PICKLOSERS)/) { # boolean
        $self->configError("$$keyPtr must be an boolean value") unless $$valuePtr =~ /^(1|y|yes|t|true|0|n|no|f|false|)$/i;
        $self->{$$keyPtr} = $$valuePtr =~ /(1|y|t)/i ? 1 : 0;
    } elsif ($$keyPtr =~ m/(PICKTYPE)/) { # boolean
        $self->configError("$$keyPtr must be either 'random' or 'first'") unless $$valuePtr =~ /^(random|first)$/i;
        $self->{$$keyPtr} = uc(substr($$valuePtr, 0, 1));
    } else { # plain old string?
        $self->{$$keyPtr} = $$valuePtr;
    }
    return 1;
}

sub loadConfig { my $self = shift;
    my $file = shift;

    return 0 unless -e $file;

    $self->{configFile} = $file;
    $self->{configKeyLine} = 0;

    open (my $fh, $file) or $self->die("Can't open config $file: $!");
    my ($key, $value) = (0, 0);
    my $EOF = '';
    my $line = 0;
    while ($_ = <$fh>) {
        $line++;
        if ($EOF) {
            if (m/^\Q$EOF\E$/) {
                $EOF = '';
                chomp $value;
                $self->applyConfig(\$key, \$value);
            } else {
                $value .= $_;
            }
        } else {
            next if (/^\s*#/ or /^\s*$/);
            s/\s+$//;
            chomp;
            $self->{configKeyLine} = $line;
            m/^\s*(\S+)\s*=\s*(.*)/ or $self->configError("invalid syntax - $_");
            ($key, $value) = ($1, $2);
            if ($value =~ /^<<\s*(\S+)/) {
                $EOF = $1;
                $value = '';
                next;
            }
            if ($key !~ /(dataTypes|dataRules)/ and $value =~ /^(['"])/) {
                my $q = $1;
                $value =~ s/^$q(.*)$q$/$1/;
            }
            $self->applyConfig(\$key, \$value);
        }
    }
    if ($EOF) {
        chomp $value;
        $self->applyConfig(\$key, \$value);
    }

    $self->{ME} = $self->{CONTESTNAME} ? "$self->{CONTESTNAME} <$self->{CONTESTEMAIL}>" : $self->{CONTESTEMAIL};
    return 1;
}

sub loadContest { my $self = shift;
#FIXME: switch log files for each contest?
    $self->{CONTEST} = shift;
    $self->loadConfig($self->contestPath('conf'));
    return 1;
}

sub subValues { my $self = shift;
    my ($strPtr, $dynamic) = @_;
    if ($dynamic) {
        while ($$strPtr =~ m/\{([^{}]+)\}/s) {
            my $cond = $1;
            my $replace = "<?>$cond<?>";
            if ($cond =~ /^(.*)\?\?(.*)\:\:(.*)$/s) {
                my ($test, $pass, $fail) = ($1, $2, $3);
                if ($test =~/^(.+):([=<>!]{1,2}):(.+)$/s) {
                    my ($a, $b) = ($1, $3);
                    $test = $2;
                    $self->subValues(\$a);
                    $self->subValues(\$b);
#FIXME: check for numeric tests?
                    if ($test eq '=') {
                        $test = ($a eq $b);
                    } elsif ($test eq '>') {
                        $test = ($a gt $b);
                    } elsif ($test =~ m/(>=|=>)/) {
                        $test = ($a ge $b);
                    } elsif ($test eq '<') {
                        $test = ($a lt $b);
                    } elsif ($test =~ m/(<=|=<)/) {
                        $test = ($a le $b);
                    } else {
                        $test = ($a ne $b);
                    }
                } else {
                    $self->subValues(\$test);
                }
                $replace = $test ? $pass : $fail;
            }
            $$strPtr =~ s/\{\Q$cond\E\}/$replace/s;
        }
    }
    my @foundKeys = $$strPtr =~m/\<([^><]+)\>/g;
    foreach (@foundKeys) {
        $_ = uc($_);
        next unless defined $self->{$_};
        $$strPtr =~ s/\<$_\>/$self->{$_}/g;
    }
}

sub mergeResults { my $self = shift;
    my $data = shift;
    foreach (keys %{$self->{dataTypes_hash}}) {
        $data->{$_} = '' unless $data->{$_};
        $self->{"RESULTS.$_"} = $data->{$_};
    }

    $self->{NOTIFYLOSUR} = $self->{'RESULTS.NOTIFY'} ? 1 : 0;
    $self->{TO} = $self->{'RESULTS.NAME'} ? "$self->{'RESULTS.NAME'} <$self->{EMAIL}>" : $self->{EMAIL};
    return;
}

sub contestRunning { my $self = shift;
    my $response = $self->{RESPONSETIME};
    my $cutoff = $self->{'EXPIRE.SECONDS'} + ($response * 60 * 60);
    if ($self->{PICKTYPE} eq 'R' and time < $cutoff) {
        my $text = "Contest Expire Date: " . $self->{'EXPIRE.DATE'} . ', ' . $self->{'EXPIRE.TIME'};
        $text .= " (plus the response time of $response hour" . ($response == 1 ? '' : 's') . ")" if ($response);
        return $text;
    }
#FIXME: do more testing for a first entry contest type
    return;
}

sub announcePending { my $self = shift;
    if (($self->{'EXPIRE.SECONDS'} + ($self->{RESPONSETIME} * 60 * 60)) > $self->{'ANNOUNCE.SECONDS'}) {
        my $runDate = $self->contestRunning;
        return $runDate if ($runDate);
    }
    my $cutoff = $self->{'ANNOUNCE.SECONDS'};
    return "Contest Anounce date: " . $self->{'ANNOUNCE.DATE'} . ', ' . $self->{'ANNOUNCE.TIME'} if (time < $cutoff);
    return '';
}

sub notify { my $self = shift;
    my ($type, $subject) = @_;
    $subject = ($subject ? $subject . ' ' : '') . "entry: $self->{IP}, $self->{EMAIL}";
    $self->writeContestLog($type, $subject);

    $type = "notify$type";
    my $to = $self->{$type} or return 0;

    $self->writeContestLog('NOTIFY', "send note to $to");

    my $time = scalar(localtime);
    my $msg = "$time\n\n";
    if ($type =~ m/(Challenge|Entered)/) {
        my $from = $self->{FROM} eq $self->{CONTESTSOURCE} ? 'Contest Entry Source' : $self->{FROM};
        $msg .= "From: $from\nIP: $self->{IP}   Email: $self->{EMAIL}\n\nResults:\n";
        foreach (sort keys %{$self->{dataTypes_hash}}) {
            $msg .= '  ' . $self->{"RESULTS.$_.DESC"} . ': ';
            my $value = $self->{"RESULTS.$_"};
            if ($self->{dataTypes_hash}->{$_} =~ /checkbox/) {
                $value = $value ? 'Checked' : 'Not Checked';
            }
            $msg .= "$value\n";
        }
        $msg .= "\n";
        main::sendMail($self, {To => $to, Subject => "($type) $subject"}, \$msg);
    } else {
        $msg .= "$self->{RAWHEADER}\n";
        $msg .= $self->{RAWBODY} if ($self->{RAWBODY});
        main::sendMail($self, {To => $to, Subject => "($type) $subject"}, \$msg, *STDIN);
    }
    return 1;
}

sub reply { my $self = shift;
    my ($type, $postSubject) = @_;
    if (my $msg = $self->{"message$type"}) {
        $self->subValues(\$msg, 1);

        my $subject = $self->{"subject$type"};
        $self->subValues(\$subject, 1);
        $subject .= " $postSubject" if ($postSubject);

#print "To: $self->{EMAIL}\nSubject: $subject\n$msg\n";
        main::sendMail($self, {To => $self->{EMAIL}, Subject => $subject}, \$msg);
    }
    return;
}

## DB ##
package Database;
use DB_File;
use Data::Dumper;

sub open {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my ($transaction, $file, $withIndex, $excludeIndex) = @_;

    $file = $transaction->contestPath($file);
    my $dbFile = "$file.db";

    my $self = {
            transaction => $transaction,
            db => 0,
            dbIndex => 0
        };

    # basic hash db
    my %db;
    tie(%db, 'DB_File', "$file.db", O_RDWR|O_CREAT, 0600)
        or $transaction->die("Can't open DB_File $file.db: $!");
    $self->{db} = \%db;


    if ($withIndex) {
        # open sorted index as well
        $DB_BTREE->{'compare'} = sub {
            my ($key1, $key2) = @_ ;
            "\L$key1" cmp "\L$key2";
        };
        my %dbIndex;
        if (tie(%dbIndex, 'DB_File', "$file.index.db", O_RDWR|O_CREAT, 0600, $DB_BTREE)) {
            $self->{dbIndex} = \%dbIndex;
            $self->{dbIndexExclude} = $excludeIndex || '';
        } else {
            my $result = "Can't open DB_File $file.index.db: $!";
            untie %db;
            $transaction->die($result);
        }
    }
    return bless $self, $class;
}

sub close { my $self = shift;
    foreach ('dbIndex', 'db') {
        next unless $self->{$_};
        my $hash = $self->{$_};
        untie %$hash;
    }
}

sub eachDB { my $self = shift;
    return each %{$self->{db}} unless $self->{dbIndex};
    (my ($indexKey, $key) = each %{$self->{dbIndex}}) or return;
    return $key, $self->{db}->{$key};
}

sub eachIndex { my $self = shift;
    return unless $self->{dbIndex};
    (my ($indexKey, $key) = each %{$self->{dbIndex}}) or return;
    my ($time) = split(/:/, $indexKey, 2);
    return $indexKey, $time, $key;
}

sub setValue { my $self = shift;
    my ($key, $value) = @_;
    $self->{db}->{$key} = $value;
}

sub value { my $self = shift;
    my $key = shift;
    return $self->{db}->{$key};
}

# ($seconds, $status, $email, $ip, $dataStr)
sub data { my $self = shift;
    my $valuePtr = shift;
    return split(/:/, $$valuePtr, 5);
}

sub delete { my $self = shift;
    my $key = shift;
    return unless exists($self->{db}->{$key});

    if ($self->{dbIndex}) {
        delete $self->{dbIndex}->{$self->indexKey($key,\$self->{db}->{$key})};
    }

    my ($seconds, $status, $email, $ip, $dataStr) = $self->data(\$self->{db}->{$key});

    delete $self->{db}->{$key};
    return ($seconds, $status, $email, $ip, $dataStr);
}

sub writeMerge { my $self = shift;
    my $key = shift;

    # this shouldn't happen very often unless the DB gets mucked about
    if ($self->{db}->{$key}) {
        my $count = 0;
        $key .= '.';
        while ($self->{db}->{$key.$count}) {
            $count++;
        }
        $key .= $count;
    }

    my $indexKey = $self->updateMerge($key);
    $self->{dbIndex}->{$indexKey} = $key if ($self->{dbIndex});
    return $key;
}

sub updateMerge { my $self = shift;
    my $key = shift;
    my $transaction = $self->{transaction};
    my $data = {};
    foreach (keys %{$transaction->{dataTypes_hash}}) {
        $data->{$_} = $transaction->{"RESULTS.$_"} || '';
    }
    $data->{'RESULTS.NOTIFY'} = $self->{NOTIFYLOSUR} unless $data->{'RESULTS.NOTIFY'};
    my ($time, $status, $email, $ip) = ($transaction->{SECONDS}, $transaction->{STATUS}, $transaction->{EMAIL}, $transaction->{IP});
    foreach ($time, $status, $email, $ip) {
        s/:+/./g;
    }
    $self->{db}->{$key} = join(':',  $time, $status, $email, $ip, Dumper($data));
    return $time.':'.$key;
}

sub getMerge { my $self = shift;
    my $key = shift;
    return unless $self->{db}->{$key};

    my $transaction = $self->{transaction};
    my ($seconds, $status, $email, $ip, $dataStr) = $self->data(\$self->{db}->{$key});
    ($transaction->{STATUS}, $transaction->{EMAIL}, $transaction->{IP}) = ($status, $email, $ip);
    $transaction->sec2date($seconds);

    my $VAR1;
    eval $dataStr;
    $self->{transaction}->mergeResults($VAR1);
    return ($seconds, $status, $email, $ip, $dataStr);
}

sub empty { my $self = shift;
    my $testMode = shift || 0;
    my $testStr = shift || '';
    my $transaction = $self->{transaction};
    if ($testMode) {
        $transaction->writeLog("${testStr}empty db");
        return unless $self->{dbIndex};
        $transaction->writeLog("${testStr}empty dbIndex");
        return;
    }
    $transaction->writeLog('empty db');
    my $hash = $self->{db};
    %$hash = ();
    return unless $self->{dbIndex};
    $transaction->writeLog('empty dbIndex');
    $hash = $self->{dbIndex};
    %$hash = ();
    return;
}

sub purgeIndexByTime { my $self = shift;
    return 0 unless $self->{dbIndex};
    my $cutTime = shift;

    my $transaction = $self->{transaction};
    my @indexKeys = ();
    while (my ($indexKey, $key) = each %{$self->{dbIndex}}) {
        my ($time) = split(/:/, $indexKey, 2);
        last unless $time < $cutTime;
        $transaction->writeContestLog('PURGE', "$cutTime, $time - $key");
        push @indexKeys, $indexKey;
    }
    my $count = @indexKeys;
    if ($count) {
        foreach my $indexKey (@indexKeys) {
            my $key = $self->{dbIndex}->{$indexKey} or return 0;
            delete $self->{db}->{$key};
            delete $self->{dbIndex}->{$indexKey};
        }
        $transaction->writeContestLog('PURGE', "purged $count pending entr" . ($count == 1 ? 'y' : 'ies'));
    }
    return $count;
}

sub indexConsistency { my $self = shift;
    return 0 unless $self->{dbIndex};
    my $testMode = shift || 0;
    my $testStr = shift || '';
    my $failedStr = $testStr . shift ||'';
    my $index = $self->{dbIndex};
    my $db = $self->{db};
    my $transaction = $self->{transaction};
    my $errors = 0;
    my $excludePatern = $self->{dbIndexExclude};
    while (my ($key, $value) = each %$db) {
        next if ($key =~ m/^$excludePatern$/);
        my $indexKey = $self->indexKey($key, \$value);
        next if exists($index->{$indexKey});
        $transaction->writeLog("\n$failedStr") unless $errors;
        $index->{$indexKey} = $key unless $testMode;
        $transaction->writeLog("$testStr  missing index: $indexKey (created in index)");
        ++$errors;
    }

    my @delete = ();
    while (my ($indexKey, $time, $key) = $self->eachIndex) {
        next if ($key !~ m/^$excludePatern$/ and exists($db->{$key}) and ($indexKey eq $self->indexKey($key, \$db->{$key})));
        $transaction->writeLog("\n$failedStr") unless $errors;
        $transaction->writeLog("$testStr  missing entry: $indexKey (removed from index)");
        push(@delete, $indexKey) unless $testMode;
        ++$errors;
    }
    foreach (@delete) {
        delete $index->{$_};
    }
    return $errors;
}

sub indexKey { my $self = shift;
    my ($key, $valuePtr) = @_;
    my ($time) = split(/:/, $$valuePtr, 2);
    return $time.':'.$key;
}

## Pending Queue ##
package Pendingdb;
use vars qw(@ISA);
@ISA = ("Database");

sub open {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $transaction = shift;
    my $self = Database->open($transaction, 'pending', 1);
    return bless $self, $class;
}

sub extract { my $self = shift;
    my $key = shift;
    my $transaction = $self->{transaction};

    # weed expired pendings
    $self->purgeIndexByTime(time() - ($transaction->{RESPONSETIME} * 60 * 60));

    # get the current key data
    my @results = $self->getMerge($key) or return;

    # remove from queue
    $self->delete($key);

    return @results;
}

## Entry Queue ##
package Entrydb;
use vars qw(@ISA);
@ISA = ("Database");

sub existsEmail { my $transaction = shift;
    my $dbEmail = Database->open($transaction, 'entry.email');
    my $self = {
            transaction => $transaction,
            dbEmail => $dbEmail
        };
    my ($emailCount, $limit) = emailValue2count($self, \$self->{dbEmail}->value(shift));
    $dbEmail->close;
    return $emailCount, $limit;
}

sub open {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $transaction = shift;
    my $self = Database->open($transaction, 'entry', 1, '(total|unique|true|winners)Count|winners');
    $self->{dbEmail} = Database->open($transaction, 'entry.email');
    $self->{totalCount} = $self->value('totalCount') || 0;
    $self->{uniqueCount} = $self->value('uniqueCount') || 0;
    $self->{trueCount} = $self->value('trueCount') || 0;
    return bless $self, $class;
}

sub DESTROY { my $self = shift;
    $self->close;
}

sub close { my $self = shift;
    $self->{dbEmail}->close if ($self->{dbEmail});
    $self->{dbEmail} = 0;
    $self->SUPER::close;
}

sub emailValue2count { my $self = shift;
    my $valuePtr = shift;

    my ($count, $limit) = (0,0);
    ($count, $limit) = split(/ /, $$valuePtr) if ($valuePtr and $$valuePtr);
    if ($count) {
        $limit = $limit || 0;
    } else {
        $count = 0;
        $limit = $self->{transaction}->{ENTRYLIMIT};
    }
    return $count + 0, $limit + 0;
}

sub eachEmail { my $self = shift; return $self->{dbEmail}->eachDB; }
sub setEmail { my $self = shift; return $self->{dbEmail}->setValue(@_); }

sub counts { my $self = shift; return $self->{totalCount}, $self->{uniqueCount}, $self->{trueCount} }
sub setEmailCount { my $self = shift;
    my ($email, $count, $limit) = @_;
    $self->{dbEmail}->setValue($email, "$count $limit");
    return;
}
sub setTotalCount { my $self = shift;
    my $count = shift;
    $self->{totalCount} = $count;
    $self->setValue('totalCount', $count);
    return;
}
sub setUniqueCount { my $self = shift;
    my $count = shift;
    $self->{uniqueCount} = $count;
    $self->setValue('uniqueCount', $count);
    return;
}
sub setTrueCount { my $self = shift;
    my $count = shift;
    $self->{trueCount} = $count;
    $self->setValue('trueCount', $count);
    return;
}
sub winners { my $self = shift;
    my @winners = ();
    @winners = split(/ /, $self->{winners}) if ($self->{winners});
    return \@winners;
}
sub setWinners { my $self = shift;
    my $winnersPtr = shift;
    $self->setValue('winners', join(' ', @$winnersPtr));
    my $count = @$winnersPtr;
    $self->setValue('winnersCount', $count);
    return;
}

sub delete { my $self = shift;
    my $key = shift;

    my ($seconds, $status, @rest) = $self->SUPER::delete($key) or return;

    my ($email) = split(/:/, $key);
    my ($emailCount, $limit) = $self->emailValue2count(\$self->{dbEmail}->value($email));
    if ($emailCount > 1) {
        $self->setEmailCount($email, --$emailCount, $limit);
    } else {
        $self->setValue('uniqueCount', --$self->{uniqueCount}) if ($self->{uniqueCount});
        $self->{dbEmail}->delete($email);
    }

    $self->setValue('totalCount', --$self->{totalCount}) if ($self->{totalCount});
    $self->setValue('trueCount', --$self->{trueCount}) if ($self->{trueCount} and $self->{transaction}->isTrue($status));
    return ($seconds, $status, @rest);
}

sub writeMerge { my $self = shift;
    my $email = shift;
    my ($emailCount, $limit) = $self->emailValue2count(\$self->{dbEmail}->value($email));

    $emailCount++;
    my $key = $self->SUPER::writeMerge("$email:$emailCount");
    $self->setEmailCount($email, $emailCount, $limit);

    $self->setValue('uniqueCount', ++$self->{uniqueCount}) if ($emailCount == 1);
    $self->setValue('totalCount', ++$self->{totalCount});
    $self->setValue('trueCount', ++$self->{trueCount}) if ($self->{transaction}->isTrue());
    return $key;
}

sub empty { my $self = shift;
    $self->SUPER::empty(@_);
    $self->{dbEmail}->empty(@_);
    return;
}

package main;

## mail ##
sub sendMail { my $transaction = shift;
    my ($parms, $msgPtr, $fh) = @_;

    $parms->{From} = $transaction->{ME} unless $parms->{From};
    $parms->{ReplyTo} = $transaction->{ME} unless $parms->{ReplyTo};
    $parms->{To} = $transaction->{TO} unless $parms->{To};
    $parms->{Subject} = '(no subject)' unless $parms->{Subject};

    eval {
        my $mailer = Mail::Mailer->new('sendmail', "-f $transaction->{CONTESTEMAIL}");
        $mailer->open($parms);
        print $mailer $$msgPtr;
        if (defined $fh) {
            while ($_ = <$fh>) {
                print $mailer $_;
            }
        }
        $mailer->close();
    };
    $transaction->die("Couldn't send mail: $@") if ($@);
}

#### PIPE MODE ####
sub pipeMode { my $transaction = shift;

    $/ = ''; # paragraph mode, read until blank line
    my $header = <STDIN>;
    $/ = "\n";
    return $transaction->cleanup unless $header;

    $transaction->{RAWHEADER} = '#' . $header;
    $header =~ s/\n\s+/ /g; # fix continuation lines

    if ($header =~ /^From\s+(\S+)/ and $1 ne '') {  # that's the Unix-style From line
        $transaction->{FROM} = $1;
    } else {
        $transaction->die("No 'From ' line!");
    }

    my $subject = $header =~ /^Subject: +(.*)$/im ? $1: '(No subject)';
    $subject =~ s/\s+$//;

    if (! ignoreEmail($transaction, $subject)) {
        if ($subject =~ /^contest #(\d+) ([0-9.]+) ([\w.-]+\@.+[\w.-]+)/i) {
            newEntry($transaction, $1, $2, $3);
        } elsif ($subject =~ /Entry #(\d+)\.(\d+)\.(\S+)$/i) {
            confirmEntry($transaction, $1, $2, $3);
        } else {
            corruptEntry($transaction, "invalid subject: $subject");
        }

    }
    $transaction->cleanup;
}

sub newEntry { my ($transaction, $contest, $ip, $email) = @_;
    ($transaction->{IP}, $transaction->{EMAIL}) = ($ip, $email);

    $transaction->loadContest($contest);

    # verify source from webserver
    if ($transaction->{CONTESTSOURCE} and ($transaction->{FROM} ne $transaction->{CONTESTSOURCE})) {
        return corruptEntry($transaction, "entry didn't come from $transaction->{CONTESTSOURCE}");
    }

    # check blacklist
    return if (blacklistEntry($transaction));

    # read the message body values
    my $data = {};
    $transaction->{RAWBODY} = '';
    my $keyCount = $transaction->{dataCount};
#FIXME: use a generic key instead of the combined secs.contest.md5
    while ($keyCount and ($_ = <STDIN>)) {
        $transaction->{RAWBODY} .= $_;
        next unless (/^\s*(\w+)[\s:]+(.*)/);
        my ($k, $v) = (uc($1), $2);
        $v =~ s/\s+$//;
        next if ($data->{$k});
        next unless (my $type = $transaction->{dataTypes_hash}->{$k});
        if ($type eq 'checkbox') {
            $v = ($v =~ m/(1|y|yes|on|$k)/i) ? 1 : 0;
        }
        $data->{$k} = $v;
        $keyCount--;
    }
    $transaction->mergeResults($data);

    return if (expiredEntry($transaction));

    # queue this into the waiting list
    return unless (my $key = savePending($transaction, $data));

    $transaction->reply('Challenge', "Entry #$key");
    $transaction->notify('Challenge');

    return 1;
}

sub submitEntry { my $transaction = shift;
    my $existing = (shift) ? 'existing' : 'confirmed';

    # test the answer and set the T or F flags
    $transaction->setStatusFlags($transaction->testRules($transaction->ruleTable));

    # record data
    my $entry = Entrydb->open($transaction);
    $entry->writeMerge($transaction->{EMAIL});
    $entry->close;

    # inform user
    $transaction->reply('Entered');
    $transaction->notify('Entered', "accept $existing");
    return 1;
}

sub confirmEntry { my ($transaction, $secs, $contest, $md5) = @_;
    $transaction->loadContest($contest);

    my $key = "$secs.$contest.$md5";
    if (crypt($transaction->{magickey} . "$contest$secs", $md5) eq $md5) {

        # find this in the waiting queue
        my $queue = Pendingdb->open($transaction);
        my $result = $queue->extract($key);
        $queue->close;

        return corruptEntry($transaction, "No pending confirmation key: $key") unless $result;

        # check blacklist and expiry date
        return if (blacklistEntry($transaction) or expiredEntry($transaction));

        # add to entry queue
        submitEntry($transaction);

    } else {
        return corruptEntry($transaction, "invalid confirmation key: $key");
    }
    return 1;
}

sub ignoreEmail { my ($transaction, $subject) = @_;
    my $from = $transaction->{FROM};

    foreach (@{$transaction->{ignore}}) {
        if ($from =~ m/^$_(\@.*)?$/i) {
            $transaction->notify('Ignore', "ignored from: $from - subject: $subject");
            return 1;
        }
    }
    return 0;
}

sub corruptEntry { my ($transaction, $reason) = @_;
    $transaction->notify('Corrupt', $reason);
    return 0;
}

sub savePending { my ($transaction, $data) = @_;
    my ($count, $limit) = Entrydb::existsEmail($transaction, $transaction->{EMAIL});
    if ($count) {
        $transaction->{ENTRYLIMIT} = $limit;
        submitEntry($transaction, 1) unless limitEntry($transaction, $limit, $count);
        return 0;
    }

    my $queue = Pendingdb->open($transaction);

    my $secs = $transaction->{SECONDS} % 259200;  # 3 day window
    my $contest = $transaction->{CONTEST};
    my @salt_chars = ('a'..'z','A'..'Z','0'..'9');
    my $salt = $salt_chars[rand(62)] . $salt_chars[rand(62)];
    my $key = "$secs.$contest.".crypt("$transaction->{magickey}$contest$secs", $salt);

    $key = $queue->writeMerge($key);
    $queue->close;

    return $key;
}

#FIXME: actually do something?
sub blacklistEntry { my $transaction = shift;

#    if ($transaction->{EMAIL} !~ m/^(leah|scott)\@littlefish\.ca$/i) {
#        $transaction->writeContestLog('TEST', "aborted (only test users allowed): $transaction->{IP}, $transaction->{EMAIL}");
#        return 1;
#    }

    return 0;

    $transaction->notify('Blacklist');
    return 1;
}

sub limitEntry { my $transaction = shift;
    my ($limit, $count) = @_;

    return 0 unless ($limit and $count >= $limit);

    $transaction->reply('Limit');
    $transaction->notify('Limit');
    return 1;
}

sub expiredEntry { my $transaction = shift;

    return 0 unless ($transaction->{SECONDS} > $transaction->{'EXPIRE.SECONDS'});

    $transaction->reply('Expired');
    $transaction->notify('Expired');
    return 1;
}

#### INTERACTIVE MODE ####
sub commandMode { my $transaction = shift;

    my $testMode = $transaction->{options}->{t};
    my $testStr = $testMode ? 'test - ' : '';

    my @contests = ();

    if ($transaction->{options}->{N}) {
        foreach my $contest (split(/,/, $transaction->{options}->{N})) {
            $transaction->{CONTEST} = $contest;
            my $contestPath = $transaction->contestPath('conf');
            if (-f $contestPath) {
                push @contests, $contest;
            } else {
                $transaction->writeLog("Contest #$contest not found");
            }
        }
    } else {
#FIXME: finish looking for contest config files
        $transaction->{CONTEST} = '(.+)';
        my $regex = $transaction->contestPath('conf');
    }

    $transaction->writeLog("No Contest(s) found") unless @contests;

#FIXME: these functions modify the {options}, affecting each loop
    foreach my $contest (@contests) {
        $transaction->writeLog("Contest #$contest");
        $transaction->loadContest($contest);

        if ($transaction->{options}->{D}) {
            dumpData($transaction, $testMode, $testStr);
        }
        if ($transaction->{options}->{L}) {
            loadData($transaction, $testMode, $testStr);
        }
        if ($transaction->{options}->{C}) {
            consistencyCheck($transaction, $testMode, $testStr);
        }
        if ($transaction->{options}->{P}) {
            pendingData($transaction, $testMode, $testStr);
        }
        if ($transaction->{options}->{E} or $transaction->{options}->{e}) {
            entryData($transaction, $testMode, $testStr);
        }
        if ($transaction->{options}->{R}) {
            ruleTest($transaction, $testMode, $testStr);
        }
        if ($transaction->{options}->{W}) {
            winners($transaction, $testMode, $testStr);
        }
        if ($transaction->{options}->{n}) {
            notifyLosers($transaction, $testMode, $testStr);
        }
    }

    $transaction->cleanup;
}

sub dumpData { my ($transaction, $testMode, $testStr) = @_;
    my $dumpFile = $transaction->path($transaction->{options}->{D});
    $transaction->{options}->{D} = 0;
    $transaction->writeLog("\nDump databases\n==============");
    my $count = 0;
    my $pendingCount = 0;
    my $entryCount = 0;
    open(my $fh, ">", $dumpFile) or $transaction->die("Failed to open $dumpFile: $!");
    my $queue = Pendingdb->open($transaction);
    while (my ($key, $value) = $queue->eachDB) {
        $count++;
        $pendingCount++;
        print $fh "pending: $key -> $value\n";
    }
    $queue->close;
    my $entry = Entrydb->open($transaction);
    while (my ($key, $value) = $entry->eachDB) {
        $count++;
        $entryCount++;
        print $fh "entry: $key -> $value\n";
    }
    while (my ($key, $value) = $entry->eachEmail) {
        $count++;
        print $fh "email: $key -> $value\n";
    }
    $entry->close;
    close($fh);
    $transaction->writeLog("\nwrote $count record" . ($count == 1 ? '' : 's'). " to file:  $dumpFile");
    $transaction->writeLog("  $pendingCount pending record" . ($pendingCount == 1 ? '' : 's'));
    $transaction->writeLog("  $entryCount entry record" . ($entryCount == 1 ? '' : 's'));
    $count -= $pendingCount + $entryCount;
    $transaction->writeLog("  $count email record" . ($count == 1 ? '' : 's'));
}

sub loadData { my ($transaction, $testMode, $testStr) = @_;
    my $loadFile = $transaction->path($transaction->{options}->{L});
    $transaction->{options}->{L} = 0;
    $transaction->writeLog("\nLoad databases\n==============");
    my $count = 0;
    my $pendingCount = 0;
    my $entryCount = 0;
    my $trueCount = 0;
    open(my $fh, $loadFile) or $transaction->die("Failed to open $loadFile: $!");
    my $queue = Pendingdb->open($transaction);
    my $entry = Entrydb->open($transaction);
    $transaction->writeLog("\nEmpty Pending database");
    $queue->empty($testMode, $testStr);
    $transaction->writeLog("Empty Entry/Email databases");
    $entry->empty($testMode, $testStr);

    my @pendingKeys = ();
    my $first = "\n";
    while (<$fh>) {
        chomp;
        my ($key, $value) = split(/ -> /);
        if ($key =~ s/^(pending|entry|email): (.+)/$2/) {
            $count++;
            if ($1 eq 'pending') {
                $pendingCount++;
                $queue->setValue($2, $value) unless $testMode;
            } elsif ($1 eq 'entry') {
                $entryCount++;
                my ($seconds, $status) = $entry->data(\$value);
                $trueCount++ if ($transaction->isTrue($status));
                $entry->setValue($2, $value) unless $testMode;
            } else {
                $entry->setEmail($2, $value) unless $testMode;
            }
        } else {
            $transaction->writeLog($first ."ignored: $_");
            $first = '';
        }
    }
    if ($testMode) {
        if ($pendingCount) {
            $transaction->writeLog("\n$testStr" . 'Rebuild pending Indexes');
            $transaction->writeLog("${testStr}Added $pendingCount Index Entr" . ($pendingCount == 1 ? 'y' : 'ies'));
        }
        if ($entryCount) {
            $transaction->writeLog("\n$testStr" . 'Rebuild entry Indexes');
            $transaction->writeLog("${testStr}Added $entryCount Index Entr" . ($entryCount == 1 ? 'y' : 'ies'));
        }
    } else {
        $entry->setTotalCount($entryCount);
        $entry->setUniqueCount($count - $entryCount - $pendingCount);
        $entry->setTrueCount($trueCount);

        my $errors = $queue->indexConsistency($testMode, $testStr, 'Rebuild pending Indexes');
        if ($errors) {
            $transaction->writeLog("${testStr}Added $errors Index Entr" . ($errors == 1 ? 'y' : 'ies'));
        }
        $errors = $entry->indexConsistency($testMode, $testStr, 'Rebuild entry Indexes');
        if ($errors) {
            $transaction->writeLog("${testStr}Added $errors Index Entr" . ($errors == 1 ? 'y' : 'ies'));
        }
    }
    $queue->close;
    $entry->close;
    close($fh);
    $transaction->writeLog("\nread $count record" . ($count == 1 ? '' : 's'). " from file:  $loadFile");
    $transaction->writeLog("  $pendingCount pending record" . ($pendingCount == 1 ? '' : 's'));
    $transaction->writeLog("  $entryCount entry record" . ($entryCount == 1 ? '' : 's'));
    $count -= $pendingCount + $entryCount;
    $transaction->writeLog("  $count email record" . ($count == 1 ? '' : 's'));
    $transaction->{options}->{C} = 1 unless $testMode;
}

sub consistencyCheck { my ($transaction, $testMode, $testStr) = @_;
    $transaction->{options}->{C} = 0;
    $transaction->writeLog("\nConsistency Check\n=================");

    my $queue = Pendingdb->open($transaction);
    # Consistency check pending database
    # look for missing index entries
    # look for missing pending entries
    my $errors = $queue->indexConsistency($testMode, $testStr, 'Pending Index Consistency check FAILED');
    if ($errors) {
        $transaction->writeLog("\nFound $errors Index Consistency Error" . ($errors == 1 ? '' : 's'));
    } else {
        $transaction->writeLog("\nPending Index Consistency check OK");
    }
    $queue->close;

    my $entry = Entrydb->open($transaction);
#FIXME consistency check!
    # Consistency check Entry and Email database
    # check email addresses not in Entries
    # check entries not in email addresses
    # check total count matches
    # check unique count matches
    # check email counts match
    $errors = $entry->indexConsistency($testMode, $testStr, 'Entry Index Consistency check FAILED');
    if ($errors) {
        $transaction->writeLog("${testStr}Found $errors Index Consistency Error" . ($errors == 1 ? '' : 's'));
    } else {
        $transaction->writeLog("\nEntry Index Consistency check OK");
    }
    $entry->close;

    ruleTest($transaction, $testMode, $testStr);
}

sub pendingData{ my ($transaction, $testMode, $testStr) = @_;
    $transaction->{options}->{P} = 0;
    $transaction->writeLog("\nPending Queue\n=============");
    my $queue = Pendingdb->open($transaction);

    my $count = 0;
    my $first = "\nPending:";
    while (my ($key, $value) = $queue->eachDB) {
        $count++;
        $transaction->writeLog("  $key -> $value");
        $first = '';
    }
    if ($count) {
        $transaction->writeLog("\n$count entr" . ($count == 1 ? 'y' : 'ies') . ' pending');
    } else {
        $transaction->writeLog("\nno entries pending");
    }

    $queue->close;
}

sub entryData{ my ($transaction, $testMode, $testStr) = @_;
    $transaction->writeLog("\nEntry Queue\n=============");
    my $entry = Entrydb->open($transaction);
    my ($totalCount, $uniqueCount, $trueCount) = $entry->counts;
    $transaction->writeLog("\nTotal Entries: $totalCount\nUnique Emails: $uniqueCount");

    my $Ecount = 0;
    if ($transaction->{options}->{E}) {
        $transaction->{options}->{E} = 0;
        if ($totalCount) {
            $transaction->writeLog("\nEntries:");
            while (my ($key, $value) = $entry->eachDB) {
                $Ecount++;
                $transaction->writeLog(" $Ecount) $value");
            }
        } else {
            $transaction->writeLog("\nno entries confirmed");
        }
    }
    my $ecount = 0;
    if ($transaction->{options}->{e}) {
        $transaction->{options}->{e} = 0;
        if ($totalCount) {
            $transaction->writeLog("\nEmail Addresses:");
            while (my ($key, $value) = $entry->eachEmail) {
                $ecount++;
                my ($count, $limit) = $entry->emailValue2count(\$value);
                $limit = $limit ? $limit : 'unlimited';
                $transaction->writeLog(" $ecount) $key - $count entr" . ($count == 1 ? 'y' : 'ies') . ", limit: $limit");
            }
        } else {
            $transaction->writeLog("\nno Email addresses confirmed");
        }
    }
    $entry->close;
}

sub ruleTest{ my ($transaction, $testMode, $testStr) = @_;
    $transaction->{options}->{R} = 0;
    $transaction->writeLog("\nTest entries against rules\n==========================");
    my $first = "\n";
    my $rulesPtr = $transaction->ruleTable;
    my $entry = Entrydb->open($transaction);
    my $count = 0;
    my ($totalCount, $uniqueCount, $trueCount) = $entry->counts;
    my $newTrueCount = 0;
    while (my ($key, $value) = $entry->eachDB) {
        $entry->getMerge($key);
        my $newStatus = $transaction->testRules($rulesPtr);
        my $prevStatus = $transaction->setStatusFlags($newStatus);
        if ($prevStatus) {
            $transaction->writeLog("$first${testStr}change status from '$prevStatus' to '$newStatus': $key");
            $entry->updateMerge($key) unless $testMode;
            $first = '';
            $count++;
        }
        ++$newTrueCount if ($transaction->isTrue);
    }
    if ($count) {
        $transaction->writeLog("  ${testStr}changed status on $count entry record" . ($count == 1 ? '' : 's'));
    } else {
        $transaction->writeLog($first.'  no status changes');
        $first = '';
    }
    if ($newTrueCount != $trueCount) {
        $transaction->writeLog("$first  ${testStr}entry 'true' count changed from $trueCount to $newTrueCount");
        $entry->setTrueCount($newTrueCount) unless $testMode;
    }
    $entry->close;
}

sub winners{ my ($transaction, $testMode, $testStr) = @_;
    $transaction->{options}->{W} = 0;
    $transaction->writeLog("\nDetermine Winner(s)\n===================");

    my $winners = $transaction->{WINNERS};
    my $contestRunning = $transaction->contestRunning;
    if ($contestRunning) {
        $transaction->writeLog("\n${testStr}The Contest winner" .
                               ($winners == 1 ? '' : 's') . " can not be determined until after the $contestRunning");
        return;
    }
    my $first = "\n";
    my $entry = Entrydb->open($transaction);
    my ($totalCount, $uniqueCount, $trueCount) = $entry->counts;
    my $failCount = $totalCount - $trueCount;
    my $winningKeys = $entry->winners;
    my $currentWinners = @$winningKeys;
#FIXME: do something more with this
    $winningKeys = ();
    my $win = winnerList($winners, $trueCount, $transaction->{PICKTYPE});

    my $losers = (!$contestRunning and ($trueCount < $winners) and $transaction->{PICKLOSERS}) ? $winners - $trueCount : 0;
    my $lose = winnerList($losers, $failCount, $transaction->{PICKTYPE});

    my $nextWin = shift @$win;
    my $nextLose = shift @$lose;
    my $newStatus;
    while (my ($key, $value) = $entry->eachDB) {
        $entry->getMerge($key);
        if ($transaction->isTrue) {
            $newStatus = winnerTest($trueCount, $nextWin, $win, $key, $winningKeys);
            $trueCount--;
        } else {
            $newStatus = winnerTest($failCount, $nextLose, $lose, $key, $winningKeys);
            $failCount--;
        }
        my $prevStatus = $transaction->setStatusFlags($newStatus);
        if ($prevStatus) {
#FIXME: check for notify status?
            $transaction->writeLog("$first${testStr}change status from '$prevStatus' to '$newStatus': $key") unless $prevStatus eq '?';
            $entry->updateMerge($key) unless $testMode;
            $first = '';
        }
    }
    $entry->close;

    foreach (@$winningKeys) {
        $transaction->writeLog($first."winning entry: $_");
        $first = '';
    }
    $winners -= @$winningKeys;
    if ($winners) {
        $transaction->writeLog("$first  $winners winning entr" . ($winners == 1 ? 'y' : 'ies') . " remain unclaimed");
        $first = '';
    }
    $transaction->writeLog("$first${testStr}update winners record");
    $entry->setWinners($winningKeys) unless $testMode;
}

sub notifyLosers{ my ($transaction, $testMode, $testStr) = @_;
    $transaction->{options}->{n} = 0;
    $transaction->writeLog("\nNotify loser(s)\n===============");
    my $announcePending = $transaction->announcePending;
    if ($announcePending) {
        $transaction->writeLog("\n${testStr}Announcements can not be determined until after the $announcePending");
        return;
    }
    my $first = "\n";
    my $notified = 0;
    my $losers = 0;
    my $entry = Entrydb->open($transaction);
    while (my ($key, $value) = $entry->eachDB) {
        $entry->getMerge($key);
        next if ($transaction->isWinner);
        if ($transaction->{NOTIFYLOSUR}) {
            if ($transaction->isNotified) {
                $losers++;
            } else {
                $transaction->writeLog("$first${testStr}send mail to $transaction->{EMAIL}: $key");
                $transaction->reply('Loser') unless $testMode;
                $transaction->setNotified;
                $entry->updateMerge($key) unless $testMode;
                $notified++;
                $first = '';
            }
        }
    }
    $entry->close;
    if ($notified) {
        $transaction->writeLog("${testStr}   notified $notified new losing entr" . ($notified == 1 ? 'y' : 'ies'));
    }
    if ($losers) {
        $transaction->writeLog("${testStr}   $losers losing entr" . ($losers == 1 ? 'y' : 'ies') . ' have already been notified');
    }

}

# ($rec, $next, $list, $key, $winningKeysPtr)
#FIXME: I don't like the hard coded 'W' and 'L' here!
sub winnerTest {
    if ($_[1] and $_[1] == $_[0]) {
        $_[1] = shift @{$_[2]};
        push @{$_[4]}, $_[3];
        return 'W';
    }
    return 'L';
}
sub winnerList { my ($require, $count, $type) = @_;
    my @sel = ();
    if ($require) {
        if ($count <= $require) {
            while ($count) {
                push @sel, $count--;
            }
        } elsif ($type eq 'R') { # type is R or F
            my @list = ();
            while ($count) {
                push @list, $count--;
            }
            while ($require) {
                my $i = rand(@list);
                push @sel, $list[$i];
                splice(@list, $i, 1);
                $require--;
            }
            @sel = sort {$b <=> $a} @sel;
        } else {
            while ($require) {
                push @sel, $count--;
                $require--;
            }
        }
    }
    return \@sel;
}

#### MAIN ####
# parse options, and determine mode
#FIXME more options - blacklist, ignore
#FIXME more interactive commands

my $options = {};
#default options
while ($stringOptions =~ m/(.)/g) {
    $options->{$1} = '';
}
while ($flagOptions =~ m/(.)/g) {
    $options->{$1} = 0;
}

while (defined($ARGV[0]) && $ARGV[0] =~ /^-/) {
    $_ = shift;
    if (/^-([$stringOptions])(.*)/) {  # value options
        $options->{$1} = $2 ? $2 : shift;
    } elsif (s/^-([$flagOptions]+)$/$1/) { # flag options
        while (/(.)/g) {
            $options->{$1} = 1;
        }
    } else {
        die $usage;
    }
}

if (@ARGV) {  # pipe mode
    pipeMode(Transaction->new(0, $options, $ARGV[0]));
} else {    # interactive mode
    commandMode(Transaction->new(1, $options, $options->{u}));
}
exit 0;
